from typing import List

from pydantic import BaseModel

from pycamel.src.modules.core.validator import Validator


class GameBase(BaseModel):
    game_name: str


class Game(GameBase):
    game_rating: int


class Order(GameBase):
    order_count: int


class OrderGame(BaseModel):
    game: Order


class User(BaseModel):
    name: str
    games: List[Game]
    orders: OrderGame


TEST_USER = {
    "name": "John",
    "games": [
        {"game_name": "CSGO", "game_rating": 5},
        {"game_name": "Diablo", "game_rating": 5}
    ],
    "orders": {
        "game": {
            "game_name": "CSGO", "order_count": 4
        }
    }
}


def test_validation_without_validation_key():
    result = Validator(User, TEST_USER).fetch()
    for item in result:
        assert isinstance(item, User) is True


def test_validation_with_validation_key():
    result = Validator(Game, TEST_USER, 'games').fetch()
    for item in result:
        assert isinstance(item, Game) is True


def test_validation_with_received_path():
    result = Validator(Order, TEST_USER, "orders:game").fetch()
    for item in result:
        assert isinstance(item, Order)


def test_validation_error_if_something_went_wrong():
    try:
        Validator(Game, TEST_USER).fetch()
        int('This row of wrong code for case when row '
            'above did not return Assertion error'
            )
    except AssertionError:
        pass



from typing import Any

import requests

from pycamel.src.modules.core.filter import Filter
from pycamel.src.modules.response.response import CamelResponse


class Router:
    """
    Default router class that gives possibility to send requests as it
    implemented in base request lib, but with some additionally functionality.
    """
    def __init__(self, path: str, router_validation_key: str = None) -> None:
        """
        :param path: Concreate router path.
        :param router_validation_key: Key that will be used for each request
        and type of request under that route.
        """
        self.path = path
        self.router_validation_key = router_validation_key
        self.headers = {'Content-Type': 'application/json'}

        self.request_path = path
        self.request_headers = self.headers

        self._execution_method = None

    def get(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.get
        return self._fetch(*args, **kwargs)

    def post(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.post
        return self._fetch(*args, **kwargs)

    def put(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.put
        return self._fetch(*args, **kwargs)

    def patch(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.patch
        return self._fetch(*args, **kwargs)

    def delete(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.delete
        return self._fetch(*args, **kwargs)

    def add_to_path(self, parameter: str) -> 'Router':
        """
        Method adds to request path any received string.
        For example, if you need to send request to localhost/api/users/12
        all that you have to do it just .add_to_path('/12').
        :param parameter: Any string.
        :return: returns self
        """
        self.request_path += parameter
        return self

    def set_headers(self, headers: dict) -> 'Router':
        """
        Method switch default application/json header to received dict with
        headers.
        :param headers: dictionary with needed headers.
        :return: returns self
        """
        self.request_headers = headers
        return self

    def set_filters(self, filters: dict) -> 'Router':
        """
        Method receives dictionary with needed filters, transform it into string
        and add to request url. For example: base url is localhost/api/users,
        and you set filters {"page": 1}, for now your request url will be
        localhost/api/users?page=1
        :param filters: dictionary with filters
        :return: returns self
        """
        self.request_path += Filter.build_filter(filters)
        return self

    def append_header(self, header_key: str, header_value: Any) -> 'Router':
        """
        Method appends to default headers, received key and value.
        For example: default header is {'Content-Type': 'application/json'},
        after append header key-APP, value-TEST, it will look like
        {'Content-Type': 'application/json', 'APP': 'TEST'}.
        :param header_key: any string value
        :param header_value: any value
        :return: returns self
        """
        self.request_headers[header_key] = header_value
        return self

    def _clear(self) -> None:
        """
        Method updates router object to default after each fetched request.
        :return: Nothing
        """
        self.request_path = self.path
        self.request_headers = self.headers

    def _fetch(self, *args, **kwargs) -> CamelResponse:
        """
        Method receives any default values from requests lib and push them into
        execution method. After request execution it returns CamelResponse.
        :param args: Any
        :param kwargs: Any
        :return: CamelResponse
        """
        response = self._execution_method(
            url=self.request_path,
            headers=self.request_headers,
            *args,
            **kwargs
        )
        _previous_headers = self.request_headers
        self._clear()
        return CamelResponse(
            response=response,
            headers=_previous_headers,
            router_validation_key=self.router_validation_key
        )



import os

from pycamel.src.modules.routing.router_maker import RouterMaker
from pycamel.src.modules.routing.router import Router


def test_router_generation(clear_project_validation_key):
    os.environ['pc_host'] = 'https://google.com'
    maker = RouterMaker('/v1').make_router(
        route='/api/images', router_validation_key='images_array')
    assert isinstance(maker, Router) is True
    assert maker.path == 'https://google.com/v1/api/images'
    assert maker.router_validation_key == 'images_array'



import os
import pytest

from pydantic import BaseModel

import requests

from pycamel.src.modules.routing.router import Router
from pycamel.src.modules.response.response import CamelResponse


class User(BaseModel):
    id: int
    email: str
    first_name: str
    last_name: str
    avatar: str


PATH = 'https://reqres.in/api/users'


@pytest.fixture
def clear_project_validation_key():
    if 'pc_project_validation_key' in os.environ:
        del os.environ['pc_project_validation_key']


@pytest.fixture
def get_router():
    test_router = Router(PATH)
    return test_router


@pytest.fixture(scope='session')
def make_request():
    response = requests.get(PATH)
    return response


@pytest.fixture
def get_response(make_request):
    c_response = CamelResponse(
        response=make_request,
        headers={'Content-Type': 'application/json'}
    )
    return c_response



import os

from pycamel.src.modules.core.config import CamelConfig


def test_config_parameters(clear_project_validation_key):
    config = CamelConfig("http://localhost/", "computer")
    assert \
        os.environ['pc_project_validation_key'] == config.project_validation_key
    assert os.environ['pc_host'] == config.host


def test_project_key_doesnt_set_if_absent_value(clear_project_validation_key):
    CamelConfig("http://localhost/")
    assert os.getenv('pc_project_validation_key') is None



from tests.conftest import User


def test_status_code_assert(get_response):
    get_response.assert_status_code([200])


def test_wrong_status_code(get_response):
    try:
        get_response.assert_status_code([500])
        int('For case when assert error will not be triggered')
    except AssertionError:
        pass


def test_incorrect_status_code_type(get_response):
    try:
        get_response.assert_status_code(200)
        assert 1 == 2, "Value error did not trigger for row above"
    except ValueError:
        pass


def test_getting_response_as_json(get_response):
    response_data = {"id": 12, "name": "John"}
    get_response.response_data = response_data
    assert get_response.get_response_json() == response_data


def test_validation_method(get_response):
    get_response.validate(User, 'data')


def test_any_parameter_assertion(get_response):
    get_response.assert_parameter("page", 1)


def test_when_parameter_not_found_for_assertion(get_response):
    try:
        get_response.assert_parameter("TV", "LG")
        int('For case when row above did not trigger assertion')
    except AssertionError:
        pass


def test_getting_items_by_key(get_response):
    result = get_response.get_items_by_key('page')
    assert isinstance(result, list) is True
    assert result == [1]


def test_getting_absent_parameter_in_response(get_response):
    result = get_response.get_items_by_key('TV')
    assert result == []



from typing import Any, List


def search_item(input_data: Any, parameter: str):
    """
    Function for searching all needed parameters across the received object.
    If nothing was found on top level, it tries to find target on low levels
    using recursion.
    :param input_data: List or dict with items that should be checked.
    :param parameter: Target key that we try to find across the object.
    :return: Returns generator.
    """
    if isinstance(input_data, dict):
        for key, value in input_data.items():
            if key == parameter:
                yield value
            else:
                yield from search_item(value, parameter)
    elif isinstance(input_data, list):
        for item in input_data:
            yield from search_item(item, parameter)


def prepare_items(input_data: Any, parameter: str) -> List:
    """
    Function collects all found items that passed parameter condition.
    :param input_data: List or dict with items that should be checked.
    :param parameter: Target key that we try to find across the object.
    :return: Returns array of accepted condition values.
    """
    result = []
    params_iterator = search_item(input_data, parameter)
    while True:
        try:
            item = params_iterator.__next__()
            if isinstance(item, list):
                for i in item:
                    result.append(i)
            else:
                result.append(item)
        except StopIteration:
            break
    return result



import os

from typing import Optional


def search_key_processor(
        router_validation_key: str = None,
        response_validation_key: str = None
) -> Optional[str]:
    """
    Functions detects all validation keys and according to priority
    returns most the highest prioritized key.
    :param router_validation_key: Key that has been received from router
    declaration. That key responses for all routes under one router maker.
    :param response_validation_key: Key that has been received from validation
    method. It has the highest priority and applies to only one data object
    in specific case.
    :return: Returns string or None if key did not find.
    """
    env_key = os.getenv("pc_project_validation_key")
    if response_validation_key is not None:
        if response_validation_key == '':
            return None
        return response_validation_key
    elif router_validation_key is not None:
        if router_validation_key != '':
            return router_validation_key
        return None
    elif env_key:
        return os.getenv("pc_project_validation_key")
    return None



import os
import pytest

from pycamel.src.utils.search_key_processor import search_key_processor


def test_empty_key_if_nothing_was_found(clear_project_validation_key):
    validation_key = search_key_processor()
    assert validation_key is None


@pytest.mark.parametrize("router_key, response_key, expected_result", [
    ("name", None, "name"),
    (None, "surname", "surname"),
    ("name", "surname", "surname"),
    ('', "last_name", 'last_name'),
    ('middle_name', '', None),
    (None, None, 'data'),
    ('', None, None)
])
def test_key_priority(
        clear_project_validation_key,
        router_key,
        response_key,
        expected_result
):
    os.environ['pc_project_validation_key'] = 'data'
    validation_key = search_key_processor(router_key, response_key)
    assert validation_key == expected_result



from typing import List, Any

from json import JSONDecodeError

import requests

from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError
from pycamel.src.modules.core.validator import Validator

from pycamel.src.utils.searcher import search_item, prepare_items
from pycamel.src.utils.search_key_processor import search_key_processor


class CamelResponse:
    """
    Class helps to work with default request.Response class.
    By default, parse some params from an object and as a result has
    functionality for validation.
    """
    def __init__(
            self,
            response: requests.Response,
            headers: dict,
            router_validation_key: str = None
    ) -> None:
        """
        :param response: Default response from requests lib.
        :param headers: requested headers needed for displaying in __str__
        inst representation.
        :param router_validation_key: validation key that has been set for
        router and will be used as default if another keys will not be sent.
        """
        self.router_validation_key = router_validation_key
        self.response = response
        self.headers = headers
        self.validated_objects = []
        try:
            self.response_data = response.json()
        except JSONDecodeError:
            self.response_data = {}

    def assert_status_code(
            self,
            expected_status_code: List[int]
    ) -> 'CamelResponse':
        """
        Validation method for response status code. Check that status code
        in list of expected status codes.
        :param expected_status_code: Receives array of status codes.
        :return: returns self
        """
        if not isinstance(expected_status_code, List):
            raise ValueError("Expected list of integers")
        assert self.response.status_code in expected_status_code, self
        return self

    def get_response_json(self) -> dict:
        """
        Method returns raw json response without any changes.
        :return: dict
        """
        return self.response_data

    def validate(
            self,
            schema: BaseModel,
            response_validation_key: str = None
    ) -> 'CamelResponse':
        """
        Method validates that to response data we can apply pydantic schema
        that has been sent to method.
        :param schema: Pydantic schema
        :param response_validation_key: Key for getting target data from
        response data. The key has the highest priority from another.
        :return: returns self
        """
        validation_key = search_key_processor(
            self.router_validation_key, response_validation_key
        )
        try:
            self.validated_objects = Validator(
                schema, self.response_data, validation_key
            ).fetch()
            return self
        except ValidationError as validation_exception:
            raise AssertionError(
                "Received objects could not be mapped to schema"
            ) from validation_exception

    def assert_parameter(
                self,
                parameter: str,
                expected_value: Any
    ) -> 'CamelResponse':
        """
        Method for validation any parameters in body. It tries to find all
        params and compare it with expected value. Params can be parsed from
        different level of response json.
        :param parameter: Searched parameter
        :param expected_value: Expected value for it.
        :return: returns self
        """
        params_iterator = search_item(self.response_data, parameter)
        parameter_found = False
        while True:
            try:
                item = params_iterator.__next__()
                assert item == expected_value, self
                parameter_found = True
            except StopIteration:
                break
        if parameter_found is False:
            raise AssertionError(
                'Searched parameter is not presented in dictionary'
            )
        return self

    def get_items_by_key(self, parameter: str) -> List:
        """
        It is not validation method. It helps to get all values by key from
        response object with level ignore.
        :param parameter: Searched parameter.
        :return: returns list of values or empty list if nothing was found.
        """
        return prepare_items(self.response_data, parameter)

    def get_validated_objects(self) -> List:
        """
        If validation method ended with successful you can get all your
        validated objects that
        :return:
        """
        return self.validated_objects

    def __str__(self) -> str:
        """
        Sting representation of response that contains base info about request
        and response.
        :return: String info about class params.
        """
        return f"\n\nRequest sent to: {self.response.url}\n" \
               f"Response status code: {self.response.status_code}\n" \
               f"Response data: {self.response_data}\n" \
               f"Headers: {self.headers}\n" \
               f"Request time took: {self.response.elapsed}"



import os

from pycamel.src.modules.routing.router import Router


class RouterMaker:
    """
    Class for specific service path, that will generate routes for the
    service. For example: you have services data-service and image-service, so
    each of them has some routers, all that you need it is just send service
    host. For now, you can create as much as you need numbers of
    endpoints.
    """

    def __init__(self, service_host: str) -> None:
        """
        Class constructor.
        :param service_host: Receives part of url.
            For example, if main url looks like that:
            https://google.com/api-service/v1/some-endpoint
            router should have that part - api-service/v1
            After it u can easily make any endpoints on that url
        """
        self.service_host = service_host

    def _build_url(self, route: str) -> str:
        """
        Build url from all parts of received data.
        :param route: String. Example of path /some-endpoint
        :return: String. Full path.
        """
        host = os.environ.get('pc_host')
        return f"{host}{self.service_host}{route}"

    def make_router(
            self,
            route: str,
            router_validation_key: str = None
    ) -> Router:
        """
        Returns Router object according to received path.
        Example of path /some-endpoint
        As a result u will get all REST methods with full functionality
        for that entity.
        :param router_validation_key: value that will be used as validation
        key for validator method. If something has been populated for
        project key validation in config, router validation key will have
        the highest priority.
        :param route: String. Example /some-endpoint
        :return: Router object
        """
        path = self._build_url(route)
        return Router(path=path, router_validation_key=router_validation_key)



import pytest
from pycamel.src.utils.searcher import prepare_items


TEST_ITEM = {
    "time": "26.02.2022",
    "name": "SolveMe",
    "age": 92,
    "games": [
        {"game_name": "Diablo"},
        {"game_name": "CSGO"},
        {"game_name": "Dota 2"}
    ],
    "friends": [
        {"name": "Erich", "age": 42},
        {"name": "Artur", "age": 35}
    ],
    "job": {
        "status": "ACTIVE",
        "title": "AQA"
    },
    "best_numbers": [1, 7, 0, 3]
}


@pytest.mark.parametrize("search_key, expected_result", [
    ("games", [
        {'game_name': 'Diablo'},
        {'game_name': 'CSGO'},
        {'game_name': 'Dota 2'}
    ]),  # array of dictionaries
    ("age", [92, 42, 35]),  # values of keys from different levels
    ("job", [{"status": "ACTIVE", "title": "AQA"}]),  # top level dictionary
    ("title", ["AQA"]),  # string from low level
    ("best_numbers", [1, 7, 0, 3]),  # array of numbers
    ("time", ["26.02.2022"]),  # string from top level
    ("datetime", [])  # case for absent key in object
])
def test_getting_different_types_of_data(search_key, expected_result):
    result = prepare_items(TEST_ITEM, search_key)
    assert result == expected_result



from tests.conftest import PATH, User


def test_path_setter(get_router):
    get_router.add_to_path('/1')
    assert get_router.request_path == f"{PATH}/1"


def test_header_setter(get_router):
    header = {"TEST_HEADER": "APP"}
    get_router.set_headers(header)
    assert get_router.request_headers == header


def test_filter_setter(get_router):
    req_filter = {"page": 2}
    get_router.set_filters(req_filter)
    assert get_router.request_path == f"{PATH}?page=2"


def test_header_append(get_router):
    get_router.append_header("TEST_HEADER", "APP")
    assert get_router.request_headers == {
        'Content-Type': 'application/json', 'TEST_HEADER': 'APP'
    }


def test_default_get_request(get_router):
    response = get_router.get()
    response.assert_status_code([200])


def test_default_post_request(get_router):
    user_data = {
        "name": "morpheus",
        "job": "leader"
    }
    response = get_router.post(json=user_data)
    response.assert_status_code([201])


def test_default_put_request(get_router):
    user_data = {
        "name": "morpheus",
        "job": "zion resident"
    }
    response = get_router.add_to_path('/2').put(json=user_data)
    response.assert_status_code([200])


def test_default_patch_request(get_router):
    user_data = {
        "name": "morpheus",
        "job": "zion resident"
    }
    response = get_router.add_to_path('/2').patch(json=user_data)
    response.assert_status_code([200])


def test_default_delete_request(get_router):
    response = get_router.add_to_path('/2').delete()
    response.assert_status_code([204])


def test_router_clear_method(get_router):
    get_router.add_to_path('/100')
    assert get_router.request_path == f"{PATH}/100"
    get_router._clear()
    assert get_router.request_path == PATH


def test_getting_assert_error_with_wrong_schema(get_router):
    response = get_router.get()
    try:
        response.validate(User, '')
        int('For case when assertion did not trigger for row above')
    except AssertionError:
        pass


def test_getting_validated_objects(get_router):
    response = get_router.add_to_path('/1').get()
    response.validate(User, 'data')
    validated_objects = response.get_validated_objects()
    assert isinstance(*validated_objects, User) is True


def test_header_propagation_to_response_class_from_set_header(get_router):
    header = {"APP": "TEST"}
    response = get_router.set_headers(header).get()
    assert response.headers == header


def test_header_propagation_to_response_class_from_append_header(get_router):
    expected_headers = {'Content-Type': 'application/json', 'APP': 'TEST'}
    response = get_router.append_header("APP", "TEST").get()
    assert response.headers == expected_headers
